@@ -594,9 +594,11 @@ namespace cpgrid
               std::vector<std::tuple<int,int,char> >,
               std::vector<std::tuple<int,int,char,int> >,
               WellConnections>
-    createZoltanListsFromParts(const CpGrid& grid, const std::vector<cpgrid::OpmWellType> * wells,
+    createListsFromParts(const CpGrid& grid, const std::vector<cpgrid::OpmWellType> * wells,
+                               const std::unordered_map<std::string, std::set<int>>& possibleFutureConnections,
                                const double* transmissibilities, const std::vector<int>& parts,
-                               bool allowDistributedWells)
+                               bool allowDistributedWells,
+                               std::shared_ptr<cpgrid::CombinedGridWellGraph> gridAndWells)
     {
         std::vector<int> exportGlobalIds;
         std::vector<int> exportLocalIds;
@@ -634,19 +636,20 @@ namespace cpgrid
             scatterExportInformation(numExport, exportGlobalIds.data(),
                                      exportToPart.data(), 0,
                                      grid.comm());
-        std::unique_ptr<cpgrid::CombinedGridWellGraph> gridAndWells;
-        if (wells)
+        if (wells and !gridAndWells)
         {
             bool partitionIsEmpty = (grid.size(0) == 0);
             EdgeWeightMethod method{}; // We don't care which method is used, we only need the graph.
             gridAndWells.reset(new cpgrid::CombinedGridWellGraph(grid,
                                                                  wells,
+                                                                 possibleFutureConnections,
                                                                  transmissibilities,
                                                                  partitionIsEmpty,
                                                                  method));
         }
         return makeImportAndExportLists(grid, grid.comm(),
                                         wells,
+                                        possibleFutureConnections,
                                         gridAndWells.get(),
                                         root,
                                         numExport,
@@ -663,6 +666,7 @@ namespace cpgrid
                std::vector<std::tuple<int,int,char,int> >,
                WellConnections>
     vanillaPartitionGridOnRoot(const CpGrid& grid, const std::vector<cpgrid::OpmWellType> * wells,
+                               const std::unordered_map<std::string, std::set<int>>& possibleFutureConnections,
                                const double* transmissibilities, bool allowDistributedWells = false)
     {
         int root = 0;
@@ -679,8 +683,7 @@ namespace cpgrid
             initialSplit[0]=cc.size()/(initialSplit[1]*initialSplit[2]);
             partition(grid, initialSplit, numParts, parts, false, false);
         }
-        return createZoltanListsFromParts(grid, wells, transmissibilities, parts,
-                                          allowDistributedWells);
+        return createListsFromParts(grid, wells, possibleFutureConnections, transmissibilities, parts, allowDistributedWells);
     }
 #endif
 } // namespace cpgrid
