@@ -154,7 +154,8 @@ compute_cell_index(const int dims[3], int i, int j,
 static int
 checkmemory(int nz, struct processed_grid *out, int **intersections)
 {
-    int r, m, n, ok;
+    size_t r, m, n;
+    int ok;
 
     /* Ensure there is enough space to manage the (pathological) case
      * of every single cell on one side of a fault connecting to all
@@ -171,7 +172,7 @@ checkmemory(int nz, struct processed_grid *out, int **intersections)
         n += MAX(n / 2, 12 * r);
     }
 
-    ok = m == out->m;
+    ok = m == ((size_t)(out->m));
     if (! ok) {
         void *p1, *p2, *p3, *p4;
 
@@ -190,7 +191,7 @@ checkmemory(int nz, struct processed_grid *out, int **intersections)
         if (ok) { out->m = m; }
     }
 
-    if (ok && (n != out->n)) {
+    if (ok && (n != ((size_t)(out->n)))) {
         void *p1;
 
         p1 = realloc(out->face_nodes, n * sizeof *out->face_nodes);
