@@ -71,6 +71,7 @@
 #include "Geometry.hpp"
 
 #include <array>
+#include <initializer_list>
 #include <set>
 #include <vector>
 
@@ -95,6 +96,15 @@ template<int> class EntityRep;
 }
 }
 
+
+void markAndAdapt_check(Dune::CpGrid&,
+                        const std::array<int,3>&,
+                        const std::vector<int>&,
+                        Dune::CpGrid&,
+                        bool,
+                        bool,
+                        bool);
+
 void refine_and_check(const Dune::cpgrid::Geometry<3, 3>&,
                       const std::array<int, 3>&,
                       bool);
@@ -114,6 +124,12 @@ void check_global_refine(const Dune::CpGrid&,
 
 void fieldProp_check(const Dune::CpGrid& grid, Opm::EclipseGrid eclGrid, std::string deck_string);
 
+void testInactiveCellsLgrs(const std::string&,
+                           const std::vector<std::array<int,3>>&,
+                           const std::vector<std::array<int,3>>&,
+                           const std::vector<std::array<int,3>>&,
+                           const std::vector<std::string>&);
+
 namespace Dune
 {
 namespace cpgrid
@@ -133,6 +149,17 @@ class CpGridData
     friend class GlobalIdSet;
     friend class HierarchicIterator;
     friend class Dune::cpgrid::IndexSet;
+    friend class Dune::cpgrid::IdSet;
+    friend class Dune::cpgrid::LevelGlobalIdSet;
+
+    friend
+    void ::markAndAdapt_check(Dune::CpGrid&,
+                              const std::array<int,3>&,
+                              const std::vector<int>&,
+                              Dune::CpGrid&,
+                              bool,
+                              bool,
+                              bool);
 
     friend
     void ::refine_and_check(const Dune::cpgrid::Geometry<3, 3>&,
@@ -140,8 +167,8 @@ class CpGridData
                             bool);
     friend
     void ::refinePatch_and_check(const std::array<int,3>&,
-                        const std::array<int,3>&,
-                        const std::array<int,3>&);
+                                 const std::array<int,3>&,
+                                 const std::array<int,3>&);
 
     friend
     void ::refinePatch_and_check(Dune::CpGrid&,
@@ -149,13 +176,18 @@ class CpGridData
                                  const std::vector<std::array<int,3>>&,
                                  const std::vector<std::array<int,3>>&,
                                  const std::vector<std::string>&);
-    
+
     friend
     void ::check_global_refine(const Dune::CpGrid&,
                                const Dune::CpGrid&);
-
     friend
     void ::fieldProp_check(const Dune::CpGrid& grid, Opm::EclipseGrid eclGrid, std::string deck_string);
+    friend
+    void ::testInactiveCellsLgrs(const std::string&,
+                                 const std::vector<std::array<int,3>>&,
+                                 const std::vector<std::array<int,3>>&,
+                                 const std::vector<std::array<int,3>>&,
+                                 const std::vector<std::string>&);
 
 private:
     CpGridData(const CpGridData& g);
@@ -206,15 +238,6 @@ public:
             return 0;
         }
     }
-    /// Read the Sintef legacy grid format ('topogeom').
-    /// \param grid_prefix the grid name, such that topology is
-    /// found in <grid_prefix>-topo.dat etc.
-    void readSintefLegacyFormat(const std::string& grid_prefix);
-
-    /// Write the Sintef legacy grid format ('topogeom').
-    /// \param grid_prefix the grid name, such that topology will be
-    /// found in <grid_prefix>-topo.dat etc.
-    void writeSintefLegacyFormat(const std::string& grid_prefix) const;
 
     /// Read the Eclipse grid format ('grdecl').
     /// \param filename the name of the file to read.
@@ -247,9 +270,11 @@ public:
     /// \param turn_normals if true, all normals will be turned. This is intended for handling inputs with wrong orientations.
     /// \param clip_z if true, the grid will be clipped so that the top and bottom will be planar.
     /// \param pichActive Whether PINCH keyword was specified
-    std::vector<std::size_t> processEclipseFormat(const Opm::EclipseGrid* ecl_grid, Opm::EclipseState* ecl_state,
-                                                  bool periodic_extension, bool turn_normals = false, bool clip_z = false,
-                                                  bool pinchActive = true);
+    /// \return A vector of removed cells and created NNCs with transmissibilities die to PINCH item 4 all.
+    std::vector<std::size_t>
+    processEclipseFormat(const Opm::EclipseGrid* ecl_grid, Opm::EclipseState* ecl_state,
+                         bool periodic_extension, bool turn_normals = false, bool clip_z = false,
+                         bool pinchActive = true);
 #endif
 
     /// Read the Eclipse grid format ('grdecl').
@@ -280,10 +305,28 @@ public:
     /// @param [out] ijk  Cartesian index triplet
     void getIJK(int c, std::array<int,3>& ijk) const
     {
-        int gc = global_cell_[c];
-        ijk[0] = gc % logical_cartesian_size_[0];  gc /= logical_cartesian_size_[0];
-        ijk[1] = gc % logical_cartesian_size_[1];
-        ijk[2] = gc / logical_cartesian_size_[1];
+        ijk = getIJK(global_cell_[c], logical_cartesian_size_);
+    }
+
+    /// @brief Extract Cartesian index triplet (i,j,k) given an index between 0 and NXxNYxNZ -1
+    ///        where NX, NY, and NZ is the total amoung of cells in each direction x-,y-,and z- respectively.
+    ///
+    /// @param [in] idx      Integer between 0 and cells_per_dim[0]*cells_per_dim[1]*cells_per_dim[2]-1
+    /// @param [in] cells_per_dim
+    /// @return Cartesian index triplet.
+    std::array<int,3> getIJK(int idx_in_parent_cell, const std::array<int,3>& cells_per_dim) const
+    {
+        // idx = k*cells_per_dim_[0]*cells_per_dim_[1] + j*cells_per_dim_[0] + i
+        // with 0<= i < cells_per_dim_[0], 0<= j < cells_per_dim_[1], 0<= k <cells_per_dim_[2].
+        assert(cells_per_dim[0]);
+        assert(cells_per_dim[1]);
+        assert(cells_per_dim[2]);
+
+        std::array<int,3> ijk = {0,0,0};
+        ijk[0] = idx_in_parent_cell % cells_per_dim[0]; idx_in_parent_cell /= cells_per_dim[0];
+        ijk[1] = idx_in_parent_cell % cells_per_dim[1];
+        ijk[2] = idx_in_parent_cell /cells_per_dim[1];
+        return ijk;
     }
 
     /// @brief Determine if a finite amount of patches (of cells) are disjoint, namely, they do not share any corner nor face.
@@ -325,7 +368,41 @@ public:
     ///                            Last cell part of the lgr will be {endIJK_vec[<patch>][0]-1, ... ,endIJK_vec[<patch>][2]-1}.
     bool patchesShareFace(const std::vector<std::array<int,3>>& startIJK_vec, const std::vector<std::array<int,3>>& endIJK_vec) const;
 
+    int sharedFaceTag(const std::vector<std::array<int,3>>& startIJK_2Patches, const std::vector<std::array<int,3>>& endIJK_2Patches) const;
+
+    
+    /// @brief Mark entity for refinement or coarsening.
+    ///
+    /// Refinement on CpGrid is partially supported for Cartesian grids, with the keyword CARFIN.
+    /// This only works for entities of codim 0.
+    ///
+    /// @param [in] refCount   To mark the element for
+    ///                        - refinement, refCount == 1
+    ///                        - doing nothing, refCount == 0
+    ///                        - coarsening, refCount == -1 (not applicable yet)
+    /// @param [in] element    Entity<0>. Currently, an element from the GLOBAL grid (level zero).
+    /// @return true, if marking was succesfull.
+    ///         false, if marking was not possible.
+    bool mark(int refCount, const cpgrid::Entity<0>& element);
+
+    /// @brief Return refinement mark for entity.
+    ///
+    /// @return refinement mark (1 refinement, 0 doing nothing, -1 coarsening - not supported yet).
+    int getMark(const cpgrid::Entity<0>& element) const;
+
+    /// @brief Set mightVanish flags for elements that will be refined in the next adapt() call
+    ///        Need to be called after elements have been marked for refinement.
+    bool preAdapt();
+
+    /// TO DO: Documentation. Triggers the grid refinement process - Currently, returns preAdapt()
+    bool adapt();
+
+    /// @brief Clean up refinement/coarsening markers - set every element to the mark 0 which represents 'doing nothing'
+    void postAdapt();
+
 private:
+    std::array<Dune::FieldVector<double,3>,8> getReferenceRefinedCorners(int idx_in_parent_cell, const std::array<int,3>& cells_per_dim) const;
+
     /// @brief Compute amount of cells in each direction of a patch of cells. (Cartesian grid required).
     ///
     /// @param [in]  startIJK  Cartesian triplet index where the patch starts.
@@ -409,6 +486,27 @@ private:
     std::array<double,3> getAverageArr(const std::vector<std::array<double,3>>& vec) const;
 
 public:
+    /// Add doc/or remove method and replace it with better approach
+    int getGridIdx() const {
+        // Not the nicest way of checking if "this" points at the leaf grid view of a mixed grid (with coarse and refined cells).
+        // 1. When the grid has been refined at least onece, level_data_ptr_ ->size() >1. Therefore, there is a chance of "this" pointing at the leaf grid view.
+        // 2. Unfortunately, level_ is default initialized by 0. This implies, in particular, that if someone wants to check the value of
+        //    "this->level_" when "this" points at the leaf grid view of a grid that has been refined, this value is - unfortunately - equal to 0.
+        // 3. Due to 2. we need an extra bool value to distinguish between the actual level 0 grid and such a leaf grid view (with incorrect level_ == 0). For this
+        //    reason we check if child_to_parent_cells_.empty() [true for actual level 0 grid, false for the leaf grid view].
+        // --- TO BE IMPROVED ---
+        if ((level_data_ptr_ ->size() >1) && (level_ == 0) && (!child_to_parent_cells_.empty())) {
+            return level_data_ptr_->size() -1;
+        }
+        return level_;
+    }
+    /// Add doc/or remove method and replace it with better approach
+    std::vector<std::shared_ptr<Dune::cpgrid::CpGridData>> levelData() const
+    {
+        return *level_data_ptr_;
+    }
+
+
     /// @brief Refine a single cell and return a shared pointer of CpGridData type.
     ///
     /// refineSingleCell() takes a cell and refines it in a chosen amount of cells (per direction); creating the
@@ -517,6 +615,12 @@ public:
         return *local_id_set_;
     }
 
+    /// Get the global index set.
+    const cpgrid::LevelGlobalIdSet& globalIdSet() const
+    {
+        return *global_id_set_;
+    }
+
     /// The logical cartesian size of the grid.
     /// This function is not part of the Dune grid interface,
     /// and should be used with caution.
@@ -540,6 +644,12 @@ public:
     template<class DataHandle>
     void communicate(DataHandle& data, InterfaceType iftype, CommunicationDirection dir);
 
+    void computeCellPartitionType();
+
+    void computePointPartitionType();
+
+    void computeCommunicationInterfaces(int noexistingPoints);
+
     /// \brief The type of the mpi communicator.
     using MPICommunicator = CpGridDataTraits::MPICommunicator ;
     /// \brief The type of the collective communication.
@@ -712,7 +822,7 @@ private:
      * in each pillar. Note that of these some may have no volume
      * and this be inactive.
      */
-    std::array<int, 3>                logical_cartesian_size_;
+    std::array<int, 3>                logical_cartesian_size_{};
     /** @brief vector with the gobal cell index for each cell.
      *
      * Note the size of this container is determined by the
@@ -738,12 +848,14 @@ private:
     std::shared_ptr<LevelGlobalIdSet> global_id_set_;
     /** @brief The indicator of the partition type of the entities */
     std::shared_ptr<PartitionTypeIndicator> partition_type_indicator_;
+    /** Mark elements to be refined **/
+    std::vector<int> mark_;
     /** Level of the current CpGridData (0 when it's "GLOBAL", 1,2,.. for LGRs). */
     int level_{0};
     /** Copy of (CpGrid object).data_ associated with the CpGridData object. */
     std::vector<std::shared_ptr<CpGridData>>* level_data_ptr_;
     // SUITABLE FOR ALL LEVELS EXCEPT FOR LEAFVIEW
-    /** Map between level and leafview cell indices. Only cells (from that level) that appear in leafview count. */  
+    /** Map between level and leafview cell indices. Only cells (from that level) that appear in leafview count. -1 when the cell vanished.*/  
     std::vector<int> level_to_leaf_cells_; // In entry 'level cell index', we store 'leafview cell index'
     /** Parent cells and their children. Entry is {-1, {}} when cell has no children.*/ // {level LGR, {child0, child1, ...}}
     std::vector<std::tuple<int,std::vector<int>>> parent_to_children_cells_; 
@@ -752,9 +864,15 @@ private:
     // SUITABLE ONLY FOR LEAFVIEW
     /** Relation between leafview and (possible different) level(s) cell indices. */ // {level, cell index in that level}
     std::vector<std::array<int,2>> leaf_to_level_cells_;
+    /** Corner history. corner_history_[ corner index ] = {level where the corner was born, its index there }, {-1,-1} otherwise. */
+    std::vector<std::array<int,2>> corner_history_;
     // SUITABLE FOR ALL LEVELS INCLUDING LEAFVIEW
     /** Child cells and their parents. Entry is {-1,-1} when cell has no father. */ // {level parent cell, parent cell index}
     std::vector<std::array<int,2>> child_to_parent_cells_;
+    /** Level-grid or Leaf-grid cell to parent cell and refined-cell-in-parent-cell index (number between zero and total amount
+        of children per parent (cells_per_dim[0]_*cells_per_dim_[1]*cells_per_dim_[2])). Entry is -1 when cell has no father. */
+    std::vector<int> cell_to_idxInParentCell_;
+    
 
 
     /// \brief Object for collective communication operations.
