@@ -44,6 +44,12 @@
 #include "PartitionTypeIndicator.hpp"
 #include <opm/grid/cpgrid/DefaultGeometryPolicy.hpp>
 
+// To be able to test local and global ids of vertices
+void refinePatch_and_check(Dune::CpGrid&,
+                           const std::vector<std::array<int,3>>&,
+                           const std::vector<std::array<int,3>>&,
+                           const std::vector<std::array<int,3>>&,
+                           const std::vector<std::string>&);
 
 namespace Dune
 {
@@ -67,6 +73,11 @@ class Entity : public EntityRep<codim>
     friend class GlobalIdSet;
     friend class HierarchicIterator;
     friend class CpGridData;
+    friend void ::refinePatch_and_check(Dune::CpGrid&,
+                                        const std::vector<std::array<int,3>>&,
+                                        const std::vector<std::array<int,3>>&,
+                                        const std::vector<std::array<int,3>>&,
+                                        const std::vector<std::string>&);
 
 public:
     /// @brief
@@ -170,10 +181,16 @@ public:
         return true;
     }
 
-    /// @brief For now, the grid is serial and the only partitionType() is InteriorEntity.
+    /// @brief In serial run, the only partitionType() is InteriorEntity.
     ///        Only needed when distributed_data_ is not empty.
     PartitionType partitionType() const;
 
+    /// @brief For parallel run, the entity - for now - does not see the CpGrid therefore we pass a bool to make
+    ///        the entity aware of the fact the the grid has been distributed.
+    ///        Each cell inherits the partition type of its origin (either parent cell or equivalent cell in level 0).
+    ///        Only needed when distributed_data_ is not empty.
+    PartitionType partitionTypeWhenLgrs(bool) const;
+
     /// @brief Return marker object (GeometryType object) representing the reference element of the entity.
     ///        Currently, cube type for all entities (cells and vertices).
     GeometryType type() const
@@ -209,16 +226,15 @@ public:
     HierarchicIterator hend(int) const;
 
     /// \brief Returns true, if the entity has been created during the last call to adapt(). Dummy.
-    bool isNew() const
-    {
-        return false;
-    }
+    bool isNew() const;
 
-    /// \brief Returns true, if entity might disappear during the next call to adapt(). Dummy.
-    bool mightVanish() const
-    {
-        return false;
-    }
+    /// \brief Returns true, if entity might disappear during the next call to adapt().
+    ///
+    ///        Currently, returns true if the element has been marked to be refined. Therefore, it "vanishes"
+    ///        in the sense that it gets replaced by its children.
+    ///        To be modified: When supporting coarsening, mightVanish should return true for all the "siblings"
+    ///        refined cells of one refined cell that got marked to be coarsened.
+    bool mightVanish() const;
 
     /// @brief ONLY FOR CELLS (Entity<0>)
     ///        Check if the entity comes from an LGR, i.e., it has been created via refinement from coarser level.
@@ -227,7 +243,7 @@ public:
     ///        local process, which can be used to test whether it is safe to call father.
     bool hasFather() const;
 
-    /// @brief  ONLY FOR CELLS (Entity<0>). Get the father Entity, in case entity.hasFather() is true.
+    /// @brief  ONLY FOR CELLS (Entity<0>). Get the father Entity (in the level-grid the father cell was born), in case entity.hasFather() is true.
     ///
     /// @return father-entity
     Entity<0> father() const;
@@ -261,15 +277,24 @@ public:
     /// \return return true if seed is pointing to a valid entity
     bool isValid () const;
 
-    /// getOrigin()
-    /// Returns parent entity in level 0, if the entity was born in any LGR.
-    /// Otherwise, returns itself. 
+    /// \brief Returns (1) parent entity in the level-grid the parent cell was born, if the entity was born in any LGR.
+    ///                (2) if the entity has no father, and is a leaf-grid-view entity, it returns the equivalent element on the level
+    ///                    that it was born. Namely, for coarse cells on the leaf never involved in any refinement process, we get the
+    ///                    equivalent entity in the GLOBAL grid (level 0). Notice that if it's a refined cell on the leaf, it does have a
+    ///                    father, in that case, this method returns the father entity.
+    ///                (3) Otherwise, returns itself. Notice that if the element is a refined one, this method returns the father() entity.
+    ///                    For a coarse cell never involved in any refinement, if it's also a leaf-grid-cell, it returns the equivalent
+    ///                    cell from level 0. Otherwise, it returns itself, which means that its grid is level 0 and the entity itself is already
+    ///                    its "origin".
     Entity<0> getOrigin() const;
-    
-    /// \brief Get equivalent element on the level grid view for an element on the leaf grid view. 
+
+    /// \brief To be invoked only for leaf-grid-view entities. Get equivalent element on the level grid the leaf-entity was born.
     Entity<0> getLevelElem() const;
 
-    /// \brief Get Cartesian Index in the level grid view where the Entity was born. 
+    /// \brief Get equivalent element on the level grid where the entity was born, if grid = leaf-grid-view. Otherwise, return itself.
+    Entity<0> getEquivLevelElem() const;
+
+    /// \brief Get Cartesian Index in the level grid view where the Entity was born.
     int getLevelCartesianIdx() const;
 
 protected:
@@ -341,6 +366,12 @@ PartitionType Entity<codim>::partitionType() const
 {
     return pgrid_->partition_type_indicator_->getPartitionType(*this);
 }
+
+template <int codim>
+PartitionType Entity<codim>::partitionTypeWhenLgrs(bool lgrsOnDistributedGrid) const
+{
+    return pgrid_->partition_type_indicator_->getPartitionTypeWhenLgrs(*this, lgrsOnDistributedGrid);
+}
 } // namespace cpgrid
 } // namespace Dune
 
@@ -415,19 +446,19 @@ bool Entity<codim>::isValid() const
 template <int codim>
 int Entity<codim>::level() const
 {
-    // if distributed_data_ is not empty, level_data_ptr_ has size 1.
-    if ((*(pgrid_ -> level_data_ptr_)).size() == 1){
-        return 0; // when there is no refinenment, level_ is not automatically instantiated
-    }
-    if (pgrid_ == (*(pgrid_->level_data_ptr_)).back().get()) { // entity on the leafview -> get the level where it was born:
-        return pgrid_ -> leaf_to_level_cells_[this-> index()][0]; // leaf_to_level_cells_ leafIdx -> {level/LGR, cell index in LGR}
-    }
-    else {
-        return pgrid_-> level_;
-    }
+    // Parallel and LGRs:
+    // If the grid has been distributed and - after that - LGRs have been added, then level_data_ptr_
+    // points at distributed_data_ which has size > 1.
+    //
+    // Serial and LGRs:
+    // If the grid is not distributed and there LGRs have been added, level_data_ptr_ points at data_ which
+    // has size > 1.
+    //
+    // - leaf_to_level_cells_ is non-empty only on the leaf grid view of a grid that has been refined.
+    // - level_ is set equal to zero when instantiating a pgrid, and rewriten when it corresponds to a refined level grid.
+    return pgrid_->leaf_to_level_cells_.empty()? pgrid_->level_ : pgrid_->leaf_to_level_cells_[this-> index()][0];
 }
 
-
 // isLeaf()
 // - if distributed_data_ is empty: an element is a leaf <-> hbegin and hend return the same iterator. Then,
 //    *cells from level 0 (coarse grid) that are not parents, are Leaf.
@@ -439,7 +470,7 @@ int Entity<codim>::level() const
 template<int codim>
 bool Entity<codim>::isLeaf() const
 {
-    if ((pgrid_ -> parent_to_children_cells_).empty()){ // LGR cells
+    if (pgrid_ -> parent_to_children_cells_.empty()){ // LGR cells
         return true;
     }
     else {
@@ -447,6 +478,27 @@ bool Entity<codim>::isLeaf() const
     }
 }
 
+template<int codim>
+bool Entity<codim>::isNew() const
+{
+    // WIP
+    // For an Entity "to be new" means that
+    // - it has a father
+    // - it has been created in the last call of adapt()
+    // To determine that, we track the level of the Entity and
+    // check if it was marked for refinement in the previos state
+    // of current_view_data_
+    return (isLeaf() && hasFather());
+}
+
+
+template<int codim>
+bool Entity<codim>::mightVanish() const
+{
+    const auto refinementMark = pgrid_ -> getMark(*this);
+    return (refinementMark == 1);
+}
+
 template<int codim>
 bool Entity<codim>::hasFather() const
 {
@@ -462,10 +514,9 @@ template<int codim>
 Entity<0> Entity<codim>::father() const
 {
     if (this->hasFather()){
-        const int& coarse_level = pgrid_ -> child_to_parent_cells_[this->index()][0]; // currently, always 0
-        const int& parent_index = pgrid_ -> child_to_parent_cells_[this->index()][1];
-        const auto& coarse_grid = (*(pgrid_ -> level_data_ptr_))[coarse_level].get();
-        return Entity<0>( *coarse_grid, parent_index, true);
+        const int& coarser_level = pgrid_ -> child_to_parent_cells_[this->index()][0];
+        const int& parent_cell_index = pgrid_ -> child_to_parent_cells_[this->index()][1];
+        return Entity<0>( *((*(pgrid_ -> level_data_ptr_))[coarser_level].get()), parent_cell_index, true);
     }
     else{
         OPM_THROW(std::logic_error, "Entity has no father.");
@@ -478,59 +529,12 @@ Dune::cpgrid::Geometry<3,3> Dune::cpgrid::Entity<codim>::geometryInFather() cons
     if (!(this->hasFather())){
         OPM_THROW(std::logic_error, "Entity has no father.");
     }
-    else{
-        // Get IJK index of the entity.
-        std::array<int,3> eIJK;
-        // Get the amount of children cell in each direction of the parent cell of the entity (same for all parents of each LGR)
-        std::array<int,3> cells_per_dim;
-        // Get "child0" IJK in the LGR
-        std::array<int,3> child0_IJK;
-        const auto& level0_grid =  (*(pgrid_->level_data_ptr_))[0];
-        const auto& child0_Idx = std::get<1>((*level0_grid).parent_to_children_cells_[this->father().index()])[0];
-        // If pgrid_ is the leafview, go to the LGR where the entity was born to get its IJK index in the LGR and the LGR dimension.
-        if (pgrid_ == (*(pgrid_->level_data_ptr_)).back().get()) // checking if pgrid_ is the LeafView
-        {
-            const auto& lgr_grid = (*(pgrid_->level_data_ptr_))[this -> level()];
-            cells_per_dim = (*lgr_grid).cells_per_dim_;
-
-            const auto& entity_lgrIdx = pgrid_ -> leaf_to_level_cells_[this->index()][1]; // leaf_to_level_cells_[cell] = {level, index}
-            (*lgr_grid).getIJK(entity_lgrIdx, eIJK);
-            (*lgr_grid).getIJK(child0_Idx, child0_IJK);
-        }
-        else // Getting grid dimension and IJK entity index when pgrid_ is an LGR
-        {
-            pgrid_ -> getIJK(this->index(), eIJK);
-            cells_per_dim = pgrid_ -> cells_per_dim_;
-            pgrid_ -> getIJK(child0_Idx, child0_IJK);
-        }
-        // Transform the local coordinates that comes from the refinemnet in such a way that the
-        // reference element of each parent cell is the unit cube. Here, (eIJK[*]-"shift")/cells_per_dim[*]
-        // Get the local coordinates of the entity (in the reference unit cube).
-        FieldVector<double, 3> corners_in_father_reference_elem_temp[8] = {
-            // corner '0'
-            { double(eIJK[0]-child0_IJK[0])/cells_per_dim[0], double(eIJK[1]-child0_IJK[1])/cells_per_dim[1],
-              double(eIJK[2]-child0_IJK[2])/cells_per_dim[2] },
-            // corner '1'
-            { double(eIJK[0]-child0_IJK[0]+1)/cells_per_dim[0], double(eIJK[1]-child0_IJK[1])/cells_per_dim[1],
-              double(eIJK[2]-child0_IJK[2])/cells_per_dim[2] },
-            // corner '2'
-            { double(eIJK[0]-child0_IJK[0])/cells_per_dim[0], double(eIJK[1]-child0_IJK[1]+1)/cells_per_dim[1],
-              double(eIJK[2]-child0_IJK[2])/cells_per_dim[2] },
-            // corner '3'
-            { double(eIJK[0]-child0_IJK[0]+1)/cells_per_dim[0], double(eIJK[1]-child0_IJK[1]+1)/cells_per_dim[1],
-              double(eIJK[2]-child0_IJK[2])/cells_per_dim[2] },
-            // corner '4'
-            { double(eIJK[0]-child0_IJK[0])/cells_per_dim[0], double(eIJK[1]-child0_IJK[1])/cells_per_dim[1],
-              double(eIJK[2]-child0_IJK[2]+1)/cells_per_dim[2] },
-            // corner '5'
-            { double(eIJK[0]-child0_IJK[0]+1)/cells_per_dim[0], double(eIJK[1]-child0_IJK[1])/cells_per_dim[1],
-              double(eIJK[2]-child0_IJK[2]+1)/cells_per_dim[2] },
-            // corner '6'
-            { double(eIJK[0]-child0_IJK[0])/cells_per_dim[0], double(eIJK[1]-child0_IJK[1]+1)/cells_per_dim[1],
-              double(eIJK[2]-child0_IJK[2]+1)/cells_per_dim[2] },
-            // corner '7'
-            { double(eIJK[0]-child0_IJK[0]+1)/cells_per_dim[0], double(eIJK[1]-child0_IJK[1]+1)/cells_per_dim[1],
-              double(eIJK[2]-child0_IJK[2]+1)/cells_per_dim[2] }};
+    auto idx_in_parent_cell = pgrid_ -> cell_to_idxInParentCell_[this->index()];
+    if (idx_in_parent_cell !=-1) {
+        const auto& cells_per_dim =  (*(pgrid_ -> level_data_ptr_))[this->level()] -> cells_per_dim_;
+        const auto& auxArr = pgrid_ -> getReferenceRefinedCorners(idx_in_parent_cell, cells_per_dim);
+        FieldVector<double, 3> corners_in_father_reference_elem_temp[8] =
+            { auxArr[0], auxArr[1], auxArr[2], auxArr[3], auxArr[4], auxArr[5], auxArr[6], auxArr[7]};
         auto in_father_reference_elem_corners = std::make_shared<EntityVariable<cpgrid::Geometry<0, 3>, 3>>();
         EntityVariableBase<cpgrid::Geometry<0, 3>>& mutable_in_father_reference_elem_corners = *in_father_reference_elem_corners;
         // Assign the corners. Make use of the fact that pointers behave like iterators.
@@ -550,21 +554,23 @@ Dune::cpgrid::Geometry<3,3> Dune::cpgrid::Entity<codim>::geometryInFather() cons
         return Dune::cpgrid::Geometry<3,3>(center_in_father_reference_elem, volume_in_father_reference_elem,
                                            in_father_reference_elem_corners, in_father_reference_elem_corner_indices_.data());
     }
+    else {
+        OPM_THROW(std::logic_error, "Entity has no father.");
+    }
 }
 
-
 template<int codim>
 Dune::cpgrid::Entity<0> Dune::cpgrid::Entity<codim>::getOrigin() const
 {
     if (hasFather())
     {
-        return this->father(); // currently, always a level 0 entity.
+        return this->father();
     }
     if (!(pgrid_ -> leaf_to_level_cells_.empty())) // entity on the LeafGridView
-    {
-        const int& entityIdxInLevel0 = pgrid_->leaf_to_level_cells_[this->index()][1]; // leaf_to_level_cells_ [leaf idx] = {0, cell idx}
-        const auto& coarse_grid = (*(pgrid_ -> level_data_ptr_))[0].get();
-        return Dune::cpgrid::Entity<0>( *coarse_grid, entityIdxInLevel0, true);
+    {  // leaf_to_level_cells_ [leaf idx] = { level where entity was born, cell idx in that level}
+        const int& levelElem = pgrid_->leaf_to_level_cells_[this->index()][0];
+        const int& levelElemIdx = pgrid_->leaf_to_level_cells_[this->index()][1];
+        return Dune::cpgrid::Entity<0>( *((*(pgrid_ -> level_data_ptr_))[levelElem].get()), levelElemIdx, true);
     }
     else
     {
@@ -577,28 +583,39 @@ Dune::cpgrid::Entity<0> Dune::cpgrid::Entity<codim>::getLevelElem() const
 {
     // Check that the element belongs to the leaf grid view
     // This is needed to get the index of the element in the level it was born.
-    // For cells born in level 0, it is enough to do elem.getOrigin().index().
-    // For cells born in LGRs/levels>0, the level index is stored in leaf_to_level_cells_
-    // leaf_to_level_cells_ [leaf idx] = {level, cell idx}
+    // leaf_to_level_cells_ [leaf idx] = {level where the entity was born, equivalent cell idx in that level}
     if (!(pgrid_ -> leaf_to_level_cells_.empty())) // entity on the LeafGridView
     {
-        const auto entityLevel = this -> level();
         const int& entityLevelIdx = pgrid_->leaf_to_level_cells_[this->index()][1];
-        const auto& level_grid = (*(pgrid_ -> level_data_ptr_))[entityLevel].get();
-        return Dune::cpgrid::Entity<0>( *level_grid, entityLevelIdx, true);
+        return Dune::cpgrid::Entity<0>( *((*(pgrid_ -> level_data_ptr_))[this->level()].get()), entityLevelIdx, true);
     }
     else {
         throw std::invalid_argument("The entity provided does not belong to the leaf grid view. ");
     }
 }
 
+template<int codim>
+Dune::cpgrid::Entity<0> Dune::cpgrid::Entity<codim>::getEquivLevelElem() const
+{
+    // Check if the element belongs to the leaf grid view
+    // This is needed to get the index of the element in the level it was born.
+    // leaf_to_level_cells_ [leaf idx] = {level where the entity was born, equivalent cell idx in that level}
+    if (!(pgrid_ -> leaf_to_level_cells_.empty())) // entity on the LeafGridView
+    {
+        const int& entityLevelIdx = pgrid_->leaf_to_level_cells_[this->index()][1];
+        return Dune::cpgrid::Entity<0>( *((*(pgrid_ -> level_data_ptr_))[this->level()].get()), entityLevelIdx, true);
+    }
+    else {
+        return *this;
+    }
+}
+
 template<int codim>
 int Dune::cpgrid::Entity<codim>::getLevelCartesianIdx() const
 {
-    const auto entityLevel = this -> level();
-    const auto level = (*(pgrid_ -> level_data_ptr_))[entityLevel].get();
-    const auto& elemInLevel = this->getLevelElem(); // throws when the entity does not belong to the leaf grid view.
-    return level -> global_cell_[elemInLevel.index()];
+    const auto& level_data = (*(pgrid_ -> level_data_ptr_))[level()].get();
+    // getLevelElem() throws when the entity does not belong to the leaf grid view.
+    return level_data -> global_cell_[getLevelElem().index()];
 }
 
 } // namespace cpgrid
