@@ -119,15 +119,6 @@ BOOST_AUTO_TEST_CASE(cellgeom)
     // Default construction.
     Geometry g_default;
 
-    // Construction from point and volume.
-    // This is a dangerous constructor kept for backwards compatibility,
-    // these checks may be removed if constructor removed.
-    // This possibly dangerous constructor is available for the benefit of
-    // CpGrid::readSintefLegacyFormat().
-    Geometry::GlobalCoordinate c(3.0);
-    Geometry::ctype v = 8.0;
-    Geometry g_dangerous(c, v);
-
     // Construction from point, volume, points and pointindices.
     // Construction from
     // 1. center
@@ -136,8 +127,8 @@ BOOST_AUTO_TEST_CASE(cellgeom)
     // 4. corner indices (a pointer to the first element of "global_refined_cell8corners_indices_storage")
     // First a unit cube, i.e. the mapping represented is the identity.
     typedef Geometry::GlobalCoordinate GC;
-    c = GC(0.5);
-    v = 1.0;
+    GC c(0.5);
+    Geometry::ctype v = 1.0;
     GC corners[8];
     GC cor;
     for (int k = 0; k < 2; ++k) {
@@ -403,7 +394,7 @@ void refine_and_check(const cpgrid::Geometry<3, 3>& parent_geometry,
 {
     using cpgrid::DefaultGeometryPolicy;
     CpGrid refined_grid;
-    auto& child_view_data = *refined_grid.current_view_data_;
+    auto& child_view_data = *(refined_grid.currentData().back());
     cpgrid::OrientedEntityTable<0, 1>& cell_to_face = child_view_data.cell_to_face_;
     Opm::SparseTable<int>& face_to_point = child_view_data.face_to_point_;
     DefaultGeometryPolicy& geometries = child_view_data.geometry_;
@@ -440,7 +431,7 @@ void refine_and_check(const cpgrid::Geometry<3, 3>& parent_geometry,
         BOOST_CHECK(equivalent_refined_grid.size(3) == refined_grid.size(3));
 
         // Check that the points (ordering/coordinates) matches
-        auto equiv_point_iter = equivalent_refined_grid.current_view_data_
+        auto equiv_point_iter = equivalent_refined_grid.currentData().back()
             ->geometry_.geomVector<3>().begin();
         for(const auto& point: geometries.geomVector<3>())
         {
@@ -449,11 +440,11 @@ void refine_and_check(const cpgrid::Geometry<3, 3>& parent_geometry,
         }
 
         // Check that cell to point matches (Implicitly checks for correct cell ordering)
-        BOOST_CHECK(refined_grid.current_view_data_->cell_to_point_ ==
-                    equivalent_refined_grid.current_view_data_->cell_to_point_);
+        BOOST_CHECK(refined_grid.currentData().back()->cell_to_point_ ==
+                    equivalent_refined_grid.currentData().back()->cell_to_point_);
 
         // Check that the cell centers and volume match
-        auto equiv_cell_iter = equivalent_refined_grid.current_view_data_
+        auto equiv_cell_iter = equivalent_refined_grid.currentData().back()
             ->geometry_.geomVector<0>().begin();
         for(const auto& cell: geometries.geomVector<0>())
         {
@@ -596,7 +587,7 @@ void refinePatch_and_check(const std::array<int,3>&,
     std::array<int, 3> end_ijk = {3,2,3};  // then patch_dim = {3-1,2-0,3-1} ={2,2,2}
     coarse_grid.createCartesian(coarse_grid_dim, cell_sizes_new);
     // Call refinePatch()
-    coarse_grid.current_view_data_->refinePatch(cells_per_dim_patch, start_ijk, end_ijk);
+    coarse_grid.currentData().front()->refinePatch(cells_per_dim_patch, start_ijk, end_ijk);
 
 }
 BOOST_AUTO_TEST_CASE(refine_patch)
