@@ -52,6 +52,8 @@
 
 #include <sstream>
 #include <iostream>
+#include <cstdlib>
+#include <cmath>
 
 struct Fixture
 {
@@ -111,17 +113,95 @@ void refinePatch_and_check(Dune::CpGrid& coarse_grid,
                            const std::vector<std::array<int,3>>& endIJK_vec,
                            const std::vector<std::string>& lgr_name_vec)
 {
-    auto& data = coarse_grid.data_;
+    auto& data = coarse_grid.currentData();
     // Add LGRs and update grid.
-    const bool faceSharing = (*data[0]).patchesShareFace(startIJK_vec, endIJK_vec);
-    if (!faceSharing){
+    try
+    {
         coarse_grid.addLgrsUpdateLeafView(cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec);
 
         BOOST_CHECK(data.size() == startIJK_vec.size() + 2);
         BOOST_CHECK( (*data[0]).child_to_parent_cells_.empty());
-        BOOST_CHECK(coarse_grid.lgr_names_["GLOBAL"] == 0);
+        BOOST_CHECK(coarse_grid.getLgrNameToLevel().at("GLOBAL") == 0);
         const auto& all_parent_cell_indices = (*data[0]).getPatchesCells(startIJK_vec, endIJK_vec);
 
+        // GLOBAL grid
+        for (int cell = 0; cell <  data[0]-> size(0); ++cell)
+        {
+            Dune::cpgrid::Entity<0> entity = Dune::cpgrid::Entity<0>(*data[0], cell, true);
+            BOOST_CHECK( entity.hasFather() == false);
+            BOOST_CHECK_THROW(entity.father(), std::logic_error);
+            BOOST_CHECK_THROW(entity.geometryInFather(), std::logic_error);
+            BOOST_CHECK( entity.getOrigin() ==  entity);
+            BOOST_CHECK( entity.getOrigin().level() == 0);
+
+            auto it = entity.hbegin(coarse_grid.maxLevel()); // With element.level(), fails
+            auto endIt = entity.hend(coarse_grid.maxLevel());
+            const auto& [lgr, childrenList] = (*data[0]).parent_to_children_cells_[entity.index()];
+            // If it == endIt, then entity.isLeaf() true (when dristibuted_data_ is empty)
+            if (it == endIt){
+                BOOST_CHECK_EQUAL(lgr, -1);
+                BOOST_CHECK(childrenList.empty());
+                BOOST_CHECK( entity.isLeaf() == true);
+                BOOST_CHECK( entity.mightVanish() == false);
+            }
+            else{
+                BOOST_CHECK(lgr != -1);
+                // If it != endIt, then entity.isLeaf() false (when dristibuted_data_ is empty)
+                BOOST_CHECK_EQUAL( it == endIt, false);
+                BOOST_CHECK( entity.mightVanish() == true);
+                BOOST_CHECK( entity.isNew() == false);
+                BOOST_CHECK_EQUAL( entity.isLeaf(), false); // parent cells do not appear in the LeafView
+
+                double referenceElemOneParent_volume = 0.;
+                std::array<double,3> referenceElem_entity_center = {0.,0.,0.}; // Expected {.5,.5,.5}
+                for (const auto& child : childrenList) {
+                    BOOST_CHECK( child != -1);
+                    BOOST_CHECK( data[lgr]-> child_to_parent_cells_[child][0] == 0);
+                    BOOST_CHECK( data[lgr]-> child_to_parent_cells_[child][1] == entity.index());
+
+                    const auto& childElem =  Dune::cpgrid::Entity<0>(*data[lgr], child, true);
+                    const auto& childGeoInFather = childElem.geometryInFather();
+
+                    BOOST_CHECK(childElem.hasFather() == true);
+                    BOOST_CHECK(childElem.level() == lgr);
+                    referenceElemOneParent_volume += childGeoInFather.volume();
+                    for (int c = 0; c < 3; ++c)  {
+                        referenceElem_entity_center[c] += childGeoInFather.center()[c];
+                    }
+                }
+
+                double referenceElemOneParent_volume_it = 0.;
+                std::array<double,3> referenceElem_entity_center_it = {0.,0.,0.}; // Expected {.5,.5,.5}
+                for (; it != endIt; ++it)
+                {
+                    BOOST_CHECK(it ->hasFather() == true);
+                    BOOST_CHECK(it ->level() == lgr);
+                    const auto& geoInFather =  it-> geometryInFather();
+
+                    referenceElemOneParent_volume_it += geoInFather.volume();
+                    for (int c = 0; c < 3; ++c)
+                    {
+                        referenceElem_entity_center_it[c] += geoInFather.center()[c];
+                    }
+                }
+                for (int c = 0; c < 3; ++c) {
+                    referenceElem_entity_center[c] /= cells_per_dim_vec[lgr-1][0]*cells_per_dim_vec[lgr-1][1]*cells_per_dim_vec[lgr-1][2];
+                    referenceElem_entity_center_it[c] /= cells_per_dim_vec[lgr-1][0]*cells_per_dim_vec[lgr-1][1]*cells_per_dim_vec[lgr-1][2];
+                }
+                BOOST_CHECK_CLOSE(referenceElemOneParent_volume, 1, 1e-13);
+                BOOST_CHECK_CLOSE(referenceElem_entity_center[0], .5, 1e-13);
+                BOOST_CHECK_CLOSE(referenceElem_entity_center[1], .5, 1e-13);
+                BOOST_CHECK_CLOSE(referenceElem_entity_center[2], .5, 1e-13);
+
+                BOOST_CHECK_CLOSE(referenceElemOneParent_volume_it, 1, 1e-13);
+                BOOST_CHECK_CLOSE(referenceElem_entity_center_it[0], .5, 1e-13);
+                BOOST_CHECK_CLOSE(referenceElem_entity_center_it[1], .5, 1e-13);
+                BOOST_CHECK_CLOSE(referenceElem_entity_center_it[2], .5, 1e-13);
+            }
+            BOOST_CHECK( entity.level() == 0);
+        } // end-GLOBAL/LevelZero-grid
+
+
         for (long unsigned int level = 1; level < startIJK_vec.size() +1; ++level) // only 1 when there is only 1 patch
         {
             check_refinedPatch_grid(cells_per_dim_vec[level-1], startIJK_vec[level-1], endIJK_vec[level-1],
@@ -129,82 +209,24 @@ void refinePatch_and_check(Dune::CpGrid& coarse_grid,
                                     (*data[level]).geometry_.template geomVector<1>(),
                                     (*data[level]).geometry_.template geomVector<3>());
             BOOST_CHECK( (*data[level]).parent_to_children_cells_.empty());
-            BOOST_CHECK(coarse_grid.lgr_names_[lgr_name_vec[level-1]] == static_cast<int>(level));
+            BOOST_CHECK(coarse_grid.getLgrNameToLevel().at(lgr_name_vec[level-1]) == static_cast<int>(level));
 
-            const auto& patch_cells = (*data[0]).getPatchCells(startIJK_vec[level-1], endIJK_vec[level-1]);
-
-            // GLOBAL grid
-            for (int cell = 0; cell <  data[0]-> size(0); ++cell)
-            {
-                Dune::cpgrid::Entity<0> entity = Dune::cpgrid::Entity<0>((*coarse_grid.data_[0]), cell, true);
-                BOOST_CHECK( entity.hasFather() == false);
-                BOOST_CHECK_THROW(entity.father(), std::logic_error);
-                BOOST_CHECK_THROW(entity.geometryInFather(), std::logic_error);
-                BOOST_CHECK( entity.getOrigin() ==  entity);
-                BOOST_CHECK( entity.getOrigin().level() == 0);
-                auto it = entity.hbegin(coarse_grid.maxLevel());
-                auto endIt = entity.hend(coarse_grid.maxLevel());
-                const auto& [lgr, childrenList] = (*coarse_grid.data_[0]).parent_to_children_cells_[cell];
-                if (std::find(all_parent_cell_indices.begin(), all_parent_cell_indices.end(), cell) == all_parent_cell_indices.end()){
-                    BOOST_CHECK_EQUAL(lgr, -1);
-                    BOOST_CHECK(childrenList.empty());
-                    BOOST_CHECK( entity.isLeaf() == true);
-                    // If it == endIt, then entity.isLeaf() true (when dristibuted_data_ is empty)
-                    BOOST_CHECK( it == endIt);
-                }
-                else{
-                    BOOST_CHECK(lgr != -1);
-                    BOOST_CHECK(childrenList.size() > 1);
-                    for (const auto& child : childrenList) {
-                        BOOST_CHECK( child != -1);
-                        BOOST_CHECK( data[lgr]-> child_to_parent_cells_[child][0] == 0);
-                        BOOST_CHECK( data[lgr]-> child_to_parent_cells_[child][1] == cell);
-                    }
-                    BOOST_CHECK_EQUAL( entity.isLeaf(), false); // parent cells do not appear in the LeafView
-                    // If it != endIt, then entity.isLeaf() false (when dristibuted_data_ is empty)
-                    BOOST_CHECK_EQUAL( it == endIt, false);
-                    // Auxiliary int to check amount of children
-                    double referenceElemOneParent_volume = 0.;
-                    std::array<double,3> referenceElem_entity_center = {0.,0.,0.}; // Expected {.5,.5,.5}
-                    for (; it != endIt; ++it)
-                    {
-                        // Do something with the son available through it->
-                        BOOST_CHECK(it ->hasFather() == true);
-                        BOOST_CHECK(it ->level() == lgr);
-                        referenceElemOneParent_volume += it-> geometryInFather().volume();
-                        for (int c = 0; c < 3; ++c)
-                        {
-                            referenceElem_entity_center[c] += (it-> geometryInFather().center())[c];
-                        }
-                        // std::cout << it->index() << '\n';
-                    }
-                    for (int c = 0; c < 3; ++c)
-                    {
-                        referenceElem_entity_center[c]
-                            /= cells_per_dim_vec[lgr-1][0]*cells_per_dim_vec[lgr-1][1]*cells_per_dim_vec[lgr-1][2];
-                    }
-                    BOOST_CHECK_CLOSE(referenceElemOneParent_volume, 1, 1e-6);
-                    BOOST_CHECK_CLOSE(referenceElem_entity_center[0], .5, 1e-6);
-                    BOOST_CHECK_CLOSE(referenceElem_entity_center[1], .5, 1e-6);
-                    BOOST_CHECK_CLOSE(referenceElem_entity_center[2], .5, 1e-6);
-                }
-                BOOST_CHECK( entity.level() == 0);
-            }
+            auto itMin = std::min_element((data[level] -> global_cell_).begin(),  (data[level] -> global_cell_).end());
+            auto itMax = std::max_element((data[level] -> global_cell_).begin(),  (data[level] -> global_cell_).end());
+            BOOST_CHECK_EQUAL( *itMin, 0);
+            const auto& maxCartesianIdxLevel = data[level]->logical_cartesian_size_[0]*data[level]->logical_cartesian_size_[1]* data[level]->logical_cartesian_size_[2] -1;
+            BOOST_CHECK_EQUAL( *itMax, maxCartesianIdxLevel);
 
             // LGRs
             for (int cell = 0; cell <  data[level]-> size(0); ++cell)
             {
-                Dune::cpgrid::Entity<0> entity = Dune::cpgrid::Entity<0>((*coarse_grid.data_[level]), cell, true);
+                Dune::cpgrid::Entity<0> entity = Dune::cpgrid::Entity<0>(*data[level], cell, true);
                 BOOST_CHECK( entity.hasFather() == true);
                 BOOST_CHECK( entity.getOrigin() ==  entity.father());
-                BOOST_CHECK( entity.index() == (data[level] -> global_cell_[entity.index()])); // global_cell_ = {0,1,..., total cells -1}
                 BOOST_CHECK( entity.getOrigin().level() == 0);
                 BOOST_CHECK_CLOSE(entity.geometryInFather().volume(),
                                   1./(cells_per_dim_vec[level-1][0]*cells_per_dim_vec[level-1][1]*cells_per_dim_vec[level-1][2]), 1e-6);
                 BOOST_CHECK(entity.father().level() == 0);
-                // Check entity.father().index() belongs to the patch_cells (corresponding LGR parents)
-                BOOST_CHECK_EQUAL((std::find(patch_cells.begin(), patch_cells.end(),
-                                             entity.father().index()) == patch_cells.end()) , false);
                 const auto& child_to_parent = (*data[level]).child_to_parent_cells_[cell];
                 BOOST_CHECK_EQUAL( child_to_parent[0] == -1, false);
                 BOOST_CHECK_EQUAL( child_to_parent[0] == 0, true);
@@ -230,10 +252,12 @@ void refinePatch_and_check(Dune::CpGrid& coarse_grid,
             {
                 const auto& faceToPoint =  (*data[startIJK_vec.size() +1]).face_to_point_[face];
                 BOOST_CHECK(faceToPoint.size() == 4);
-                for (int i = 0; i < 4; ++i)
-                {
+                for (int i = 0; i < 4; ++i) {
                     BOOST_CHECK((*data[startIJK_vec.size() +1]).face_to_point_[face][i] != -1);
                 }
+
+                Dune::cpgrid::EntityRep<1> faceEntity(face, true);
+                BOOST_CHECK((*data[startIJK_vec.size() +1]).face_to_cell_[faceEntity].size() < 3);
             }
 
             // LeafView
@@ -244,7 +268,7 @@ void refinePatch_and_check(Dune::CpGrid& coarse_grid,
                 {
                     BOOST_CHECK( data[startIJK_vec.size()+1] -> cell_to_point_[cell][i] != -1);
                 }
-                Dune::cpgrid::Entity<0> entity = Dune::cpgrid::Entity<0>((*coarse_grid.data_[startIJK_vec.size()+1]), cell, true);
+                Dune::cpgrid::Entity<0> entity = Dune::cpgrid::Entity<0>(*data[startIJK_vec.size()+1], cell, true);
                 for (int i = 0; i < data[startIJK_vec.size()+1] -> cell_to_face_[entity].size(); ++i)
                 {
                     BOOST_CHECK( data[startIJK_vec.size()+1] -> cell_to_face_[entity][i].index() != -1);
@@ -267,8 +291,6 @@ void refinePatch_and_check(Dune::CpGrid& coarse_grid,
                     BOOST_CHECK_EQUAL( child_to_parent[0] == 0, true);
                     BOOST_CHECK_EQUAL( child_to_parent[1], entity.father().index());
                     BOOST_CHECK( entity.father() == entity.getOrigin());
-                    BOOST_CHECK(  (data[startIJK_vec.size() +1] -> global_cell_[entity.index()]) ==
-                                  (data[0] -> global_cell_[entity.getOrigin().index()]) );
                     BOOST_CHECK( entity.getOrigin().level() == 0);
                     BOOST_CHECK( std::get<0>((*data[0]).parent_to_children_cells_[child_to_parent[1]]) == entity.level());
                     BOOST_CHECK_EQUAL((std::find(std::get<1>((*data[0]).parent_to_children_cells_[child_to_parent[1]]).begin(),
@@ -437,6 +459,12 @@ void refinePatch_and_check(Dune::CpGrid& coarse_grid,
         BOOST_CHECK( static_cast<int>(startIJK_vec.size()) == coarse_grid.maxLevel());
         BOOST_CHECK( (*data[data.size()-1]).parent_to_children_cells_.empty());
 
+        auto itMin = std::min_element((data.back() -> global_cell_).begin(),  (data.back()-> global_cell_).end());
+        auto itMax = std::max_element((data.back() -> global_cell_).begin(),  (data.back() -> global_cell_).end());
+        BOOST_CHECK_EQUAL( *itMin, 0);
+        const auto& maxCartesianIdx = coarse_grid.logicalCartesianSize()[0]*coarse_grid.logicalCartesianSize()[1]*coarse_grid.logicalCartesianSize()[2] -1;
+        BOOST_CHECK_EQUAL( *itMax, maxCartesianIdx);
+
         for (long unsigned int l = 0; l < startIJK_vec.size() +1; ++l) // level 0,1,2,... , last patch
         {
             const auto& view = coarse_grid.levelGridView(l);
@@ -447,26 +475,102 @@ void refinePatch_and_check(Dune::CpGrid& coarse_grid,
 
         std::vector<int> leaf_to_parent_cell; // To store parent cell index, when leaf cell has a parent. Empty entry otherwise.
         leaf_to_parent_cell.resize(data[startIJK_vec.size()+1]-> size(0)); // Correct size.
-        //
-        const auto& leaf_view = coarse_grid.leafGridView();
-        const auto& level0_view = coarse_grid.levelGridView(0);
-        Dune::MultipleCodimMultipleGeomTypeMapper<Dune::CpGrid::LeafGridView> leafMapper(leaf_view, Dune::mcmgElementLayout());
-        Dune::MultipleCodimMultipleGeomTypeMapper<Dune::CpGrid::LevelGridView> level0Mapper(level0_view, Dune::mcmgElementLayout());
-        const auto& leaf_idSet = (*data[startIJK_vec.size()+1]).local_id_set_;
-        const auto& level0_idSet = (*data[0]).local_id_set_;
-        for (const auto& element: elements(leaf_view)){
+
+        Dune::MultipleCodimMultipleGeomTypeMapper<Dune::CpGrid::LeafGridView> leafMapper(coarse_grid.leafGridView(), Dune::mcmgElementLayout());
+        Dune::MultipleCodimMultipleGeomTypeMapper<Dune::CpGrid::LevelGridView> level0Mapper(coarse_grid.levelGridView(0), Dune::mcmgElementLayout());
+
+        for (const auto& element: elements(coarse_grid.leafGridView())){
             BOOST_CHECK( ((element.level() >= 0) || (element.level() < static_cast<int>(startIJK_vec.size()) +1)));
             if (element.hasFather()) { // leaf_cell has a father!
                 leaf_to_parent_cell[leafMapper.index(element)] = level0Mapper.index(element.father());
-                const auto& id = (*leaf_idSet).id(element);
-                const auto& parent_id = (*level0_idSet).id(element.father());
-                BOOST_CHECK(element.index() == id);
+                const auto& parent_id = data[0]->localIdSet().id(element.father());
                 BOOST_CHECK(element.index() == leafMapper.index(element));
                 BOOST_CHECK(element.father().index() == leaf_to_parent_cell[element.index()]);
                 BOOST_CHECK(element.father().index() == parent_id);
                 BOOST_CHECK(element.father().index() == level0Mapper.index(element.father()));
             }
         }
+
+        std::set<int> allIds_set;
+        std::vector<int> allIds_vec;
+        allIds_vec.reserve(data.back()->size(0) + data.back()->size(3));
+        for (const auto& element: elements(coarse_grid.leafGridView())){
+            const auto& localId = data.back()->localIdSet().id(element);
+            const auto& globalId = data.back()->globalIdSet().id(element);
+            // In serial run, local and global id coincide:
+            BOOST_CHECK_EQUAL(localId, globalId);
+            allIds_set.insert(localId);
+            allIds_vec.push_back(localId);
+            // Check that the global_id_set_ptr_ has the correct id (id from the level where the entity was born).
+            BOOST_CHECK_EQUAL( coarse_grid.globalIdSet().id(element), data[element.level()]->localIdSet().id(element.getEquivLevelElem()));
+        }
+        // Check injectivity of the map local_id_set_ (and, indirectly, global_id_set_) after adding cell ids.
+        BOOST_CHECK( allIds_set.size() == allIds_vec.size());
+
+        for (const auto& point : vertices(coarse_grid.leafGridView())){
+            const auto& localId = data.back()->localIdSet().id(point);
+            const auto& globalId = data.back()->globalIdSet().id(point);
+            BOOST_CHECK_EQUAL(localId, globalId);
+            allIds_set.insert(localId);
+            allIds_vec.push_back(localId);
+            // Check that the global_id_set_ptr_ has the correct id (id from the level where the entity was born).
+        }
+        // Check injectivity of the map local_id_set_ (and, indirectly, global_id_set_) after adding point ids.
+        BOOST_CHECK( allIds_set.size() == allIds_vec.size());
+
+        // Local/Global id sets for level grids (level 0, 1, ..., maxLevel)
+        for (int level = 0; level < coarse_grid.maxLevel() +1; ++level)
+        {
+            std::set<int> levelIds_set;
+            std::vector<int> levelIds_vec;
+            levelIds_vec.reserve(data[level]->size(0) + data[level]->size(3));
+            for (const auto& element: elements(coarse_grid.levelGridView(level))){
+                const auto& localId = data[level]->localIdSet().id(element);
+                const auto& globalId = data[level]->globalIdSet().id(element);
+                // In serial run, local and global id coincide:
+                BOOST_CHECK_EQUAL(localId, globalId);
+                levelIds_set.insert(localId);
+                levelIds_vec.push_back(localId);
+                // Search in the leaf grid view elements for the element with the same id, if it exists.
+                if (auto itIsLeaf = std::find_if( elements(coarse_grid.leafGridView()).begin(),
+                                                  elements(coarse_grid.leafGridView()).end(),
+                                                  [localId, data](const Dune::cpgrid::Entity<0>& leafElem)
+                                                  { return (localId == data.back()->localIdSet().id(leafElem)); });
+                    itIsLeaf != elements(coarse_grid.leafGridView()).end()) {
+                    BOOST_CHECK( itIsLeaf->getEquivLevelElem() == element);
+                }
+                if (element.isLeaf()) { // Check that the id of a cell not involved in any further refinement appears on the IdSet of the leaf grid view.
+                    BOOST_CHECK( std::find(allIds_set.begin(), allIds_set.end(), localId) != allIds_set.end());
+                }
+                else { // Check that the id of a cell that vanished during refinement does not appear on the IdSet of the leaf grid view.
+                    BOOST_CHECK( std::find(allIds_set.begin(), allIds_set.end(), localId) == allIds_set.end());
+                }
+                const auto& idx = data[level]->indexSet().index(element);
+                // In serial run, local and global id coincide:
+                BOOST_CHECK_EQUAL(idx, element.index());
+            }
+
+            for (const auto& point : vertices(coarse_grid.levelGridView(level))) {
+                const auto& localId = data[level]->localIdSet().id(point);
+                const auto& globalId = data[level]->globalIdSet().id(point);
+                BOOST_CHECK_EQUAL(localId, globalId);
+                levelIds_set.insert(localId);
+                levelIds_vec.push_back(localId);
+                // Search in the leaf grid view elements for the element with the same id, if it exists.
+                if (auto itIsLeaf = std::find_if( vertices(coarse_grid.leafGridView()).begin(),
+                                                  vertices(coarse_grid.leafGridView()).end(),
+                                                  [localId, data](const Dune::cpgrid::Entity<3>& leafPoint)
+                                                  { return (localId == data.back()->localIdSet().id(leafPoint)); });
+                    itIsLeaf != vertices(coarse_grid.leafGridView()).end()) {
+                    BOOST_CHECK( (*itIsLeaf).geometry().center() == point.geometry().center() );
+                }
+            }
+            // Check injectivity of the map local_id_set_ (and, indirectly, global_id_set_)
+            BOOST_CHECK( levelIds_set.size() == levelIds_vec.size());
+        }
+    }
+    catch (const std::exception& e) {
+        std::cout << e.what() << "'\n";
     }
 }
 
@@ -483,8 +587,7 @@ BOOST_AUTO_TEST_CASE(refine_patch_different_cell_sizes)
     const std::string lgr_name = {"LGR1"};
     coarse_grid.createCartesian(grid_dim, cell_sizes);
     refinePatch_and_check(coarse_grid, {cells_per_dim}, {startIJK}, {endIJK}, {lgr_name});
-    BOOST_CHECK_EQUAL(coarse_grid.chooseData()[0]->patchesShareFace({startIJK}, {endIJK}), false);
-    std::cout << "Patches do not share faces." << "\n";
+    BOOST_CHECK_EQUAL(coarse_grid.currentData()[0]->patchesShareFace({startIJK}, {endIJK}), false);
 }
 
 BOOST_AUTO_TEST_CASE(refine_patch)
@@ -499,8 +602,7 @@ BOOST_AUTO_TEST_CASE(refine_patch)
     const std::string lgr_name = {"LGR1"};
     coarse_grid.createCartesian(grid_dim, cell_sizes);
     refinePatch_and_check(coarse_grid, {cells_per_dim}, {startIJK}, {endIJK}, {lgr_name});
-    BOOST_CHECK_EQUAL(coarse_grid.chooseData()[0]->patchesShareFace({startIJK}, {endIJK}), false);
-    std::cout << "Patches do not share faces." << "\n";
+    BOOST_CHECK_EQUAL(coarse_grid.currentData()[0]->patchesShareFace({startIJK}, {endIJK}), false);
 }
 
 BOOST_AUTO_TEST_CASE(refine_patch_one_cell)
@@ -515,8 +617,7 @@ BOOST_AUTO_TEST_CASE(refine_patch_one_cell)
     const std::string lgr_name = {"LGR1"};
     coarse_grid.createCartesian(grid_dim, cell_sizes);
     refinePatch_and_check(coarse_grid, {cells_per_dim}, {startIJK}, {endIJK}, {lgr_name});
-    BOOST_CHECK_EQUAL(coarse_grid.chooseData()[0]->patchesShareFace({startIJK}, {endIJK}), false);
-    std::cout << "Patches do not share faces." << "\n";
+    BOOST_CHECK_EQUAL(coarse_grid.currentData()[0]->patchesShareFace({startIJK}, {endIJK}), false);
 }
 
 BOOST_AUTO_TEST_CASE(lgrs_disjointPatches)
@@ -531,8 +632,7 @@ BOOST_AUTO_TEST_CASE(lgrs_disjointPatches)
     const std::vector<std::array<int,3>> endIJK_vec = {{2,1,1}, {1,1,3}, {4,3,3}};
     const std::vector<std::string> lgr_name_vec = {"LGR1", "LGR2", "LGR3"};
     refinePatch_and_check(coarse_grid, cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec);
-    BOOST_CHECK_EQUAL(coarse_grid.chooseData()[0]->patchesShareFace(startIJK_vec, endIJK_vec), false);
-    std::cout << "Patches do not share faces." << "\n";
+    BOOST_CHECK_EQUAL(coarse_grid.currentData()[0]->patchesShareFace(startIJK_vec, endIJK_vec), false);
 }
 
 BOOST_AUTO_TEST_CASE(lgrs_disjointPatchesB)
@@ -547,8 +647,7 @@ BOOST_AUTO_TEST_CASE(lgrs_disjointPatchesB)
     const std::vector<std::array<int,3>> endIJK_vec = {{2,2,1}, {4,3,3}};
     const std::vector<std::string> lgr_name_vec = {"LGR1", "LGR2"};
     refinePatch_and_check(coarse_grid, cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec);
-    BOOST_CHECK_EQUAL(coarse_grid.chooseData()[0]->patchesShareFace(startIJK_vec, endIJK_vec), false);
-    std::cout << "Patches do not share faces." << "\n";
+    BOOST_CHECK_EQUAL(coarse_grid.currentData()[0]->patchesShareFace(startIJK_vec, endIJK_vec), false);
 }
 
 BOOST_AUTO_TEST_CASE(patches_share_corner)
@@ -563,7 +662,7 @@ BOOST_AUTO_TEST_CASE(patches_share_corner)
     const std::vector<std::array<int,3>> endIJK_vec = {{1,1,1}, {2,2,2}, {4,3,3}};
     const std::vector<std::string> lgr_name_vec = {"LGR1", "LGR2", "LGR3"};
     refinePatch_and_check(coarse_grid, cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec);
-    BOOST_CHECK_EQUAL(coarse_grid.chooseData()[0]->patchesShareFace(startIJK_vec, endIJK_vec), false);
+    BOOST_CHECK_EQUAL(coarse_grid.currentData()[0]->patchesShareFace(startIJK_vec, endIJK_vec), false);
 }
 
 BOOST_AUTO_TEST_CASE(patches_share_edge)
@@ -578,10 +677,10 @@ BOOST_AUTO_TEST_CASE(patches_share_edge)
     const std::vector<std::array<int,3>> endIJK_vec = {{2,1,1}, {3,1,2}, {4,3,3}};
     const std::vector<std::string> lgr_name_vec = {"LGR1", "LGR2", "LGR3"};
     refinePatch_and_check(coarse_grid, cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec);
-    BOOST_CHECK_EQUAL(coarse_grid.chooseData()[0]->patchesShareFace(startIJK_vec, endIJK_vec), false);
+    BOOST_CHECK_EQUAL(coarse_grid.currentData()[0]->patchesShareFace(startIJK_vec, endIJK_vec), false);
 }
 
-BOOST_AUTO_TEST_CASE(pathces_share_face)
+BOOST_AUTO_TEST_CASE(patches_share_face)
 {
     // Create a grid
     Dune::CpGrid coarse_grid;
@@ -592,34 +691,71 @@ BOOST_AUTO_TEST_CASE(pathces_share_face)
     const std::vector<std::array<int,3>> startIJK_vec = {{0,0,0}, {2,0,0}, {3,2,2}};
     const std::vector<std::array<int,3>> endIJK_vec = {{2,1,1}, {3,1,1}, {4,3,3}};
     const std::vector<std::string> lgr_name_vec = {"LGR1", "LGR2", "LGR3"};
-    BOOST_CHECK_THROW(coarse_grid.addLgrsUpdateLeafView(cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec), std::logic_error);
-    BOOST_CHECK_EQUAL(coarse_grid.chooseData()[0]->patchesShareFace(startIJK_vec, endIJK_vec), true);
-    std::cout << "Patches shared at least one face." << "\n";
+    refinePatch_and_check(coarse_grid, cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec);
 }
 
-BOOST_AUTO_TEST_CASE(pathces_share_faceB)
+BOOST_AUTO_TEST_CASE(patches_share_faceB)
 {
     // Create a grid
     Dune::CpGrid coarse_grid;
     const std::array<double, 3> cell_sizes = {1.0, 1.0, 1.0};
     const std::array<int, 3> grid_dim = {4,3,3};
     coarse_grid.createCartesian(grid_dim, cell_sizes);
-    const std::vector<std::array<int,3>> cells_per_dim_vec = {{2,2,2}, {2,2,2}, {2,2,2}};
-    const std::vector<std::array<int,3>> startIJK_vec = {{0,0,0}, {0,0,1}, {1,1,2}};
-    const std::vector<std::array<int,3>> endIJK_vec = {{2,2,1}, {3,2,2}, {4,3,3}};
-    const std::vector<std::string> lgr_name_vec = {"LGR1", "LGR2", "LGR3"};
-    BOOST_CHECK_THROW(coarse_grid.addLgrsUpdateLeafView(cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec), std::logic_error);
-    BOOST_CHECK_EQUAL(coarse_grid.chooseData()[0]->patchesShareFace(startIJK_vec, endIJK_vec), true);
-    std::cout << "Patches shared at least one face." << "\n";
+    const std::vector<std::array<int,3>> cells_per_dim_vec = {{4,2,2}, {3,2,2}};
+    const std::vector<std::array<int,3>> startIJK_vec = {{0,0,0}, {1,0,0}};
+    const std::vector<std::array<int,3>> endIJK_vec = {{1,1,1}, {2,1,1}};
+    const std::vector<std::string> lgr_name_vec = {"LGR1", "LGR2"};
+    refinePatch_and_check(coarse_grid, cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec);
 }
 
+BOOST_AUTO_TEST_CASE(patches_share_Iface_with_diff_cells_per_dim)
+{
+    // Create a grid
+    Dune::CpGrid coarse_grid;
+    const std::array<double, 3> cell_sizes = {1.0, 1.0, 1.0};
+    const std::array<int, 3> grid_dim = {4,3,3};
+    coarse_grid.createCartesian(grid_dim, cell_sizes);
+    const std::vector<std::array<int,3>> cells_per_dim_vec = {{4,2,5}, {3,2,2}};
+    const std::vector<std::array<int,3>> startIJK_vec = {{0,0,0}, {1,0,0}};
+    const std::vector<std::array<int,3>> endIJK_vec = {{1,1,1}, {2,1,1}};
+    const std::vector<std::string> lgr_name_vec = {"LGR1", "LGR2"};
+    refinePatch_and_check(coarse_grid, cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec);
+}
+
+BOOST_AUTO_TEST_CASE(patches_share_Jface_with_diff_cells_per_dim)
+{
+    // Create a grid
+    Dune::CpGrid coarse_grid;
+    const std::array<double, 3> cell_sizes = {1.0, 1.0, 1.0};
+    const std::array<int, 3> grid_dim = {4,3,3};
+    coarse_grid.createCartesian(grid_dim, cell_sizes);
+    const std::vector<std::array<int,3>> cells_per_dim_vec = {{4,2,5}, {3,2,2}};
+    const std::vector<std::array<int,3>> startIJK_vec = {{0,0,0}, {0,1,0}};
+    const std::vector<std::array<int,3>> endIJK_vec = {{1,1,1}, {1,2,1}};
+    const std::vector<std::string> lgr_name_vec = {"LGR1", "LGR2"};
+    refinePatch_and_check(coarse_grid, cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec);
+}
+
+BOOST_AUTO_TEST_CASE(patches_share_Kface_with_diff_cells_per_dim)
+{
+    // Create a grid
+    Dune::CpGrid coarse_grid;
+    const std::array<double, 3> cell_sizes = {1.0, 1.0, 1.0};
+    const std::array<int, 3> grid_dim = {4,3,3};
+    coarse_grid.createCartesian(grid_dim, cell_sizes);
+    const std::vector<std::array<int,3>> cells_per_dim_vec = {{4,2,5}, {3,2,2}};
+    const std::vector<std::array<int,3>> startIJK_vec = {{0,0,0}, {0,0,1}};
+    const std::vector<std::array<int,3>> endIJK_vec = {{1,1,1}, {1,1,2}};
+    const std::vector<std::string> lgr_name_vec = {"LGR1", "LGR2"};
+    refinePatch_and_check(coarse_grid, cells_per_dim_vec, startIJK_vec, endIJK_vec, lgr_name_vec);
+}
 
 
 void check_global_refine(const Dune::CpGrid& refined_grid, const Dune::CpGrid& equiv_fine_grid)
 {
 
-    const auto& refined_data = refined_grid.data_;
-    const auto& equiv_data = equiv_fine_grid.data_;
+    const auto& refined_data = refined_grid.currentData();
+    const auto& equiv_data = equiv_fine_grid.currentData();
 
     BOOST_CHECK(refined_data.size()==3);
 
@@ -629,48 +765,65 @@ void check_global_refine(const Dune::CpGrid& refined_grid, const Dune::CpGrid& e
     // Check the container sizes
     BOOST_CHECK_EQUAL(refined_leaf.face_to_cell_.size(), equiv_leaf.face_to_cell_.size());
     BOOST_CHECK_EQUAL(refined_leaf.face_to_point_.size(), equiv_leaf.face_to_point_.size());
-    BOOST_CHECK_EQUAL(refined_leaf.cell_to_point_.size(), equiv_leaf.cell_to_point_.size());
     BOOST_CHECK_EQUAL(refined_leaf.face_normals_.size(), equiv_leaf.face_normals_.size());
+    BOOST_CHECK_EQUAL(refined_leaf.face_tag_.size(), equiv_leaf.face_tag_.size());
+    BOOST_CHECK_EQUAL(refined_leaf.cell_to_point_.size(), equiv_leaf.cell_to_point_.size());
+    BOOST_CHECK_EQUAL(refined_leaf.cell_to_face_.size(), equiv_leaf.cell_to_face_.size());
+    BOOST_CHECK_EQUAL(refined_leaf.geomVector<3>().size(), equiv_leaf.geomVector<3>().size());
+
+    BOOST_CHECK_EQUAL(refined_grid.size(3), equiv_fine_grid.size(3));
+    BOOST_CHECK_EQUAL(refined_grid.size(0), equiv_fine_grid.size(0));
 
-    // Check that the points (ordering/coordinates) matches
-    auto equiv_point_iter = equiv_leaf.geomVector<3>().begin();
     for(const auto& point: refined_leaf.geomVector<3>())
     {
+        auto equiv_point_iter = equiv_leaf.geomVector<3>().begin();
+        while ((equiv_point_iter != equiv_leaf.geomVector<3>().end()) && (point.center() != equiv_point_iter->center())) {
+            ++equiv_point_iter;
+        }
         CHECK_COORDINATES(point.center(), equiv_point_iter->center());
         for(const auto& coord: point.center())
             BOOST_TEST(std::isfinite(coord));
-        ++equiv_point_iter;
+
     }
-    auto equiv_cell_iter = equiv_leaf.geomVector<3>().begin();
     for(const auto& cell: refined_leaf.geomVector<3>())
     {
+        auto equiv_cell_iter = equiv_leaf.geomVector<3>().begin();
+        while ((equiv_cell_iter != equiv_leaf.geomVector<3>().end()) && (cell.center() != equiv_cell_iter->center())) {
+            ++equiv_cell_iter;
+        }
         CHECK_COORDINATES(cell.center(), equiv_cell_iter->center());
         for(const auto& coord: cell.center())
             BOOST_TEST(std::isfinite(coord));
         BOOST_CHECK_CLOSE(cell.volume(), equiv_cell_iter->volume(), 1e-6);
-        ++equiv_cell_iter;
     }
 
     /////
     const auto& grid_view = refined_grid.leafGridView();
     const auto& equiv_grid_view = equiv_fine_grid.leafGridView();
 
-    auto equiv_element_iter = equiv_grid_view.begin<0>();
+
     for(const auto& element: elements(grid_view))
     {
         BOOST_CHECK( element.getOrigin().level() == 0);
-        for(const auto& intersection: intersections(grid_view, element))
-        {
+        auto equiv_element_iter = equiv_grid_view.begin<0>();
+        bool closedCenter =  (std::abs(element.geometry().center()[0] - equiv_element_iter->geometry().center()[0]) < 1e-12) &&
+            (std::abs(element.geometry().center()[1] - equiv_element_iter->geometry().center()[1]) < 1e-12) &&
+            (std::abs(element.geometry().center()[2] - equiv_element_iter->geometry().center()[2])< 1e-12);
+
+        while ((equiv_element_iter != equiv_grid_view.end<0>()) && (!closedCenter)) {
+            ++equiv_element_iter;
+            closedCenter = (std::abs(element.geometry().center()[0] - equiv_element_iter->geometry().center()[0]) < 1e-12) &&
+                (std::abs(element.geometry().center()[1] - equiv_element_iter->geometry().center()[1]) < 1e-12) &&
+                (std::abs(element.geometry().center()[2] - equiv_element_iter->geometry().center()[2])< 1e-12);
+        }
+        for(const auto& intersection: intersections(grid_view, element)) {
             // find matching intersection (needed as ordering is allowed to be different
             bool matching_intersection_found = false;
-            for(auto& intersection_match: intersections(equiv_grid_view, *equiv_element_iter))
-            {
-                if(intersection_match.indexInInside() == intersection.indexInInside())
-                {
+            for(auto& intersection_match: intersections(equiv_grid_view, *equiv_element_iter)) {
+                if(intersection_match.indexInInside() == intersection.indexInInside()) {
                     BOOST_CHECK(intersection_match.neighbor() == intersection.neighbor());
 
-                    if(intersection.neighbor())
-                    {
+                    if(intersection.neighbor()) {
                         BOOST_CHECK(intersection_match.indexInOutside() == intersection.indexInOutside());
                     }
 
@@ -678,14 +831,13 @@ void check_global_refine(const Dune::CpGrid& refined_grid, const Dune::CpGrid& e
                     const auto& geom_match = intersection_match.geometry();
                     BOOST_TEST(0.0 == 1e-11, boost::test_tools::tolerance(1e-8));
                     const auto& geom =  intersection.geometry();
-                    BOOST_CHECK_CLOSE(geom_match.volume(), geom.volume(), 1e-6);
+                    BOOST_CHECK_CLOSE(geom_match.volume(), geom.volume(), 1e-13);
                     CHECK_COORDINATES(geom_match.center(), geom.center());
                     BOOST_CHECK(geom_match.corners() == geom.corners());
 
                     decltype(geom.corner(0)) sum_match{}, sum{};
 
-                    for(int cor = 0; cor < geom.corners(); ++cor)
-                    {
+                    for(int cor = 0; cor < geom.corners(); ++cor) {
                         sum += geom.corner(cor);
                         sum_match += geom_match.corner(1);
                     }
@@ -693,12 +845,10 @@ void check_global_refine(const Dune::CpGrid& refined_grid, const Dune::CpGrid& e
                     matching_intersection_found = true;
                     break;
                 }
-            }
+            } // end-for-loop-intersection_match
             BOOST_CHECK(matching_intersection_found);
         }
-        ++equiv_element_iter;
     }
-    /////
 }
 
 
@@ -740,7 +890,7 @@ BOOST_AUTO_TEST_CASE(global_norefine)
     coarse_grid.createCartesian(grid_dim, cell_sizes);
     coarse_grid.addLgrsUpdateLeafView({cells_per_dim}, {startIJK}, {endIJK}, {lgr_name});
 
-    // Create a 8x6x6 grid with length 4x3x3
+    // Create a 4x3x3 grid with length 4x3x3
     Dune::CpGrid fine_grid;
     const std::array<double, 3> fine_cell_sizes = {1.0, 1.0, 1.0};
     const std::array<int, 3> fine_grid_dim = {4,3,3};
@@ -748,4 +898,3 @@ BOOST_AUTO_TEST_CASE(global_norefine)
 
     check_global_refine(coarse_grid, fine_grid);
 }
-
