@@ -75,7 +75,7 @@ BOOST_GLOBAL_FIXTURE(Fixture);
 
 void lookup_check(const Dune::CpGrid& grid)
 {
-    const auto& data = grid.data_;
+    const auto& data = grid.currentData();
 
     std::vector<int> fake_feature(data[0]->size(0), 0);
     std::iota(fake_feature.begin(), fake_feature.end(), 3);
@@ -86,8 +86,8 @@ void lookup_check(const Dune::CpGrid& grid)
     std::vector<std::vector<int>> fakeLgrFeatures;
     fakeLgrFeatures.resize(data.size()-1);
     // Creating fake field properties for each LGR
-    if (data.size()>1) {
-        for (long unsigned int lgr = 1; lgr < data.size(); ++lgr)
+    if (grid.maxLevel()>0) {
+        for (int lgr = 1; lgr < grid.maxLevel() +1; ++lgr)
         {
             std::vector<int> fake_feature_lgr(data[lgr]->size(0), lgr);
             fakeLgrFeatures[lgr-1] = fake_feature_lgr;
@@ -111,8 +111,6 @@ void lookup_check(const Dune::CpGrid& grid)
     const auto& level0_view = grid.levelGridView(0);
     const Dune::MultipleCodimMultipleGeomTypeMapper<Dune::CpGrid::LeafGridView> leafMapper(leaf_view, Dune::mcmgElementLayout());
     const Dune::MultipleCodimMultipleGeomTypeMapper<Dune::CpGrid::LevelGridView> level0Mapper(level0_view, Dune::mcmgElementLayout());
-
-    const auto& leaf_idSet = (*data.back()).localIdSet();
     const auto& level0_idSet = (*data[0]).localIdSet();
 
     for (const auto& elem : elements(leaf_view)) {
@@ -168,7 +166,7 @@ void lookup_check(const Dune::CpGrid& grid)
             // Checks for CartesianCoordinateLevel
             const auto idxOnLevel = elem.getLevelElem().index(); // getLevelElemt throws when entity does not belong to the leafGridView
             std::array<int,3> ijkLevelGrid;
-            (*grid.data_[elem.level()]).getIJK(idxOnLevel, ijkLevelGrid);
+            (*data[elem.level()]).getIJK(idxOnLevel, ijkLevelGrid);
             std::array<int,3> ijkLevel;
             cartMapper.cartesianCoordinateLevel(idxOnLevel, ijkLevel, elem.level());
             BOOST_CHECK( ijkLevelGrid == ijkLevel);
@@ -177,9 +175,7 @@ void lookup_check(const Dune::CpGrid& grid)
         BOOST_CHECK(featureInElem == level0Mapper.index(elem.getOrigin()) +3);
         BOOST_CHECK(featureInElem == fake_feature[lookUpData.getFieldPropIdx(elem)]);
         if (elem.hasFather()) { // leaf_cell has a father!
-            const auto id = leaf_idSet.id(elem);
             const auto parent_id = level0_idSet.id(elem.father());
-            BOOST_CHECK(elem.index() == id);
             BOOST_CHECK(elem.index() == leafMapper.index(elem));
             BOOST_CHECK(elem.father().index() == featureInElem -3);
             BOOST_CHECK(elem.father().index() == parent_id);
@@ -199,6 +195,7 @@ BOOST_AUTO_TEST_CASE(one_lgr_grid)
     const std::array<int, 3> cells_per_dim = {2,2,2};
     const std::array<int, 3> startIJK = {1,0,1};
     const std::array<int, 3> endIJK = {3,2,3};  // patch_dim = {3-1, 2-0, 3-1} ={2,2,2}
+    // Cells to be refined {13,14,17,18, 25,26,29,30}
     const std::string lgr_name = {"LGR1"};
     grid.addLgrsUpdateLeafView({cells_per_dim}, {startIJK}, {endIJK}, {lgr_name});
 
@@ -337,12 +334,12 @@ void fieldProp_check(const Dune::CpGrid& grid, Opm::EclipseGrid eclGrid, std::st
             const auto& parentIdx = elem.father().index();
             const auto& parentPorv = porv[parentIdx];
             // Remark: children_list indices are the indices on the LGR - Not on the leaf grid View.
-            const auto& [lgr, children_list] = (*grid.chooseData()[0]).parent_to_children_cells_[parentIdx];
+            const auto& [lgr, children_list] = (*grid.currentData()[0]).parent_to_children_cells_[parentIdx];
             // Get child indices on the leaf grid view, get their porv value, sum them up, and compare
             // the sum with the pore volume of their parent.
             double sumChildrenPorv = 0.;
             for (const auto& child : children_list) {
-                const auto& childIdxOnLeaf = (*grid.chooseData()[lgr]).level_to_leaf_cells_[child];
+                const auto& childIdxOnLeaf = (*grid.currentData()[lgr]).level_to_leaf_cells_[child];
                 sumChildrenPorv += porvOnLeaf[childIdxOnLeaf];
             }
             BOOST_CHECK_CLOSE(parentPorv, sumChildrenPorv, 1e-6);
@@ -357,7 +354,7 @@ void fieldProp_check(const Dune::CpGrid& grid, Opm::EclipseGrid eclGrid, std::st
 
 
 BOOST_AUTO_TEST_CASE(fieldProp) {
-  const std::string deckString =
+    const std::string deckString =
         R"( RUNSPEC
         DIMENS
         1  1  5 /
@@ -409,7 +406,7 @@ BOOST_AUTO_TEST_CASE(fieldProp) {
 
 
 BOOST_AUTO_TEST_CASE(fieldPropLgr) {
-  const std::string deckString = R"( RUNSPEC
+    const std::string deckString = R"( RUNSPEC
 DIMENS
 1 1 5
 /
